==> fred-backend-messenger_0.4.0.orig.tar.gz: fred-backend-messenger-0.4.0/messenger/tests/utils.py <==
"""Utilities for tests."""
import os
from copy import copy
from typing import Any, Dict, Sequence, Type, cast
from unittest import TestCase
from unittest.mock import ANY
from uuid import UUID

from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from click import BaseCommand
from click.testing import CliRunner, Result
from easysettings import YAMLSettings
from sqlalchemy import create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.event import listen, remove
from sqlalchemy.orm import Session
from sqlalchemy.pool import Pool
from sqlalchemy.sql import text
from sqlalchemy_utils.functions import (create_database as _create_database, database_exists as _database_exists,
                                        drop_database as _drop_database)
from testfixtures import LogCapture

from messenger.commands.migrate_archive import get_fred_engine
from messenger.constants import ArchiveMessageStatus, MessageStatus
from messenger.database import get_engine, get_session_cls
from messenger.filters import Filter
from messenger.models import (ArchiveEmail, ArchiveEmailReference, ArchiveLetter, ArchiveLetterReference, ArchiveSms,
                              ArchiveSmsReference, Email, Letter, Sms)
from messenger.settings import SETTINGS, Settings

try:
    from psycopg2.extensions import connection as psycopg2_connection
except ImportError:
    psycopg2_connection = type(None)


class TestSettings(Settings):
    """Settings manager for tests.

    Does not actually load settings from any configuration file.
    """

    def load(self, config_argv: str = None) -> None:
        """Initiate an empty settings.

        Does nothing if called multiple times.
        """
        if self._settings is None:
            self._settings = YAMLSettings()


def database_exists(url):
    """Check if database exists.

    Tweaked version to use template1 database to investigate postgresql databases and use environment variables.
    """
    url = copy(make_url(url))
    if url.drivername == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if opts:
            database = url.database
        else:
            # No connection parameters, use environment instead.
            database = os.environ.get('PGDATABASE')
        url = url.set(database='template1')
        sql = text("SELECT 1 FROM pg_database WHERE datname= :database")

        engine = create_engine(url)
        try:
            return bool(engine.scalar(sql, database=database))
        finally:
            engine.dispose()
    return _database_exists(url)


def create_database(url, encoding='utf-8', template=None):
    """Create database.

    Tweaked version to use postgresql environment variables, when appropriate.
    """
    if url.drivername == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if not opts:
            # No connection parameters, use environment instead.
            url = url.set(database=os.environ.get('PGDATABASE'))
    return _create_database(url, encoding=encoding, template=template)


def drop_database(url):
    """Drop database.

    Tweaked version to use postgresql environment variables, when appropriate.
    """
    if url.drivername == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if not opts:
            # No connection parameters, use environment instead.
            url = url.set(database=os.environ.get('PGDATABASE'))
    return _drop_database(url)


def run_migrations(engine):
    """Run migrations on target database engine."""
    connection = engine.connect()
    config = Config(os.path.join(os.path.dirname(__file__), '..', 'alembic.ini'))
    scripts = ScriptDirectory.from_config(config)
    context = EnvironmentContext(config, scripts)

    def _upgrade(revision, context):
        return scripts._upgrade_revs(scripts.get_heads(), revision)

    context.configure(connection=connection, fn=_upgrade)
    context.run_migrations()
    if engine.name == 'postgresql':
        sql = "CREATE TABLE IF NOT EXISTS {table}_default PARTITION OF {table} DEFAULT"
        # Create default partitions for partitioned tables.
        engine.execute(sql.format(table=ArchiveEmail.__tablename__))
        engine.execute(sql.format(table=ArchiveEmailReference.__tablename__))
        engine.execute(sql.format(table=ArchiveSms.__tablename__))
        engine.execute(sql.format(table=ArchiveSmsReference.__tablename__))
        engine.execute(sql.format(table=ArchiveLetter.__tablename__))
        engine.execute(sql.format(table=ArchiveLetterReference.__tablename__))


def set_connection_timezone(connection, record):
    """Set time zone for postgres connection to UTC."""
    if isinstance(connection, psycopg2_connection):
        cursor = connection.cursor()
        cursor.execute('SET TIME ZONE UTC')


# XXX: Using snake_case looks much better for this context manager. It reminds override_settings from Django.
class override_settings:
    """Decorator and context manager to temporarily override settings."""

    def __init__(self, **overrides):
        self.overrides = overrides
        self.old_settings = None

    @property
    def overrides_database(self):
        """Return whether override changes database."""
        return any(n.startswith('db_') for n in self.overrides)

    def reset_database(self):
        """Reset database if required."""
        if self.overrides_database:
            get_engine.cache_clear()
            get_session_cls.cache_clear()
        if any(n.startswith('fred_db_') for n in self.overrides):
            get_fred_engine.cache_clear()

    def enable(self):
        """Start settings override."""
        self.old_settings = SETTINGS.settings
        SETTINGS.settings = TestSettings()
        # Capture logs from settings loading.
        with LogCapture('messenger.settings'):
            for key, value in dict(self.old_settings.items(), **self.overrides).items():
                SETTINGS[key] = value
        self.reset_database()
        if self.overrides_database:
            listen(Pool, 'connect', set_connection_timezone)

            # Create and migrate test database.
            if not database_exists(get_engine().url):
                create_database(get_engine().url)
            run_migrations(get_engine())

    def disable(self):
        """Stop settings override."""
        if self.overrides_database and not get_engine().name == 'sqlite':  # pragma: no cover
            # Drop the test database, if not a sqlite database.
            drop_database(get_engine().url)

            remove(Pool, 'connect', set_connection_timezone)

        SETTINGS.settings = cast(Settings, self.old_settings)
        self.reset_database()

    def __enter__(self):
        self.enable()

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()

    def __call__(self, decorated: Type[TestCase]) -> Type[TestCase]:
        """Decorate TestCase."""
        original_setup = decorated.setUp
        original_teardown = decorated.tearDown

        def setUp(inner_self):
            self.enable()
            try:
                original_setup(inner_self)
            except Exception:  # pragma: no cover
                self.disable()
                raise

        def tearDown(inner_self):
            original_teardown(inner_self)
            self.disable()

        decorated.setUp = setUp  # type: ignore
        decorated.tearDown = tearDown  # type: ignore
        return decorated


def make_email(
        session: Session, *, recipient: Sequence[str], subject_template: str, body_template: str,
        status: MessageStatus = MessageStatus.PENDING, **kwargs: Any) -> Email:
    """Make and save an email."""
    email = Email(recipient=recipient, status=status, subject_template=subject_template,
                  body_template=body_template, **kwargs)
    session.add(email)
    session.commit()
    return email


def make_archive_email(
        session: Session, *, id: int, recipient: Sequence[str], subject_template: str, body_template: str,
        status: ArchiveMessageStatus = ArchiveMessageStatus.SENT, **kwargs: Any) -> ArchiveEmail:
    """Make and save an archive email."""
    email = ArchiveEmail(id=id, recipient=recipient, status=status, subject_template=subject_template,
                         body_template=body_template, **kwargs)
    session.add(email)
    session.commit()
    return email


def make_sms(session: Session, *, recipient: str, body_template: str, status: MessageStatus = MessageStatus.PENDING,
             **kwargs: Any) -> Sms:
    """Make and save a sms."""
    sms = Sms(recipient=recipient, status=status, body_template=body_template, **kwargs)
    session.add(sms)
    session.commit()
    return sms


def make_archive_sms(session: Session, *, id: int, recipient: str, body_template: str,
                     status: ArchiveMessageStatus = ArchiveMessageStatus.SENT, **kwargs: Any) -> ArchiveSms:
    """Make and save an archive SMS."""
    sms = ArchiveSms(id=id, recipient=recipient, status=status, body_template=body_template, **kwargs)
    session.add(sms)
    session.commit()
    return sms


def make_letter(session: Session, *, file: UUID, status: MessageStatus = MessageStatus.PENDING,
                **kwargs: Any) -> Letter:
    """Make and save a letter."""
    letter = Letter(status=status, file=file, **kwargs)
    session.add(letter)
    session.commit()
    return letter


def make_archive_letter(
        session: Session, *, id: int, file: UUID, status: ArchiveMessageStatus = ArchiveMessageStatus.SENT,
        **kwargs: Any) -> ArchiveLetter:
    """Make and save an archive letter."""
    letter = ArchiveLetter(id=id, status=status, file=file, **kwargs)
    session.add(letter)
    session.commit()
    return letter


class StaticFilter(Filter):
    """Filter which always return static value.

    By default, it accept all messages, as if no filter was defined.
    """

    def __init__(self, result: bool = True):
        self.result = result

    def filter(self, model: Any) -> bool:
        # Return whatever is inserted.
        return self.result


class CommandMixin:
    """Mixin for command tests."""

    runner = CliRunner(mix_stderr=False)
    command: BaseCommand

    def invoke(self, args: Sequence[str] = None, *, settings: Dict = None) -> Result:
        """Invoke a command and return a result."""
        with override_settings(**(settings or {})):
            return self.runner.invoke(self.command, args)

    def assertCommandSuccess(self, args: Sequence[str] = None, *, stdout: str = ANY, settings: Dict = None) -> None:
        """Invoke command and check it succeeded."""
        result = self.invoke(args, settings=settings)

        err_msg = "command failed with code {}, stderr: {!r}".format(result.exit_code, result.stderr)
        cast(TestCase, self).assertEqual(result.exit_code, 0, msg=err_msg)
        cast(TestCase, self).assertEqual(result.stdout, stdout)
        cast(TestCase, self).assertEqual(result.stderr, '')

    def assertCommandFail(self, args: Sequence[str] = None, *, stdout: str = '', stderr: str = '',
                          settings: Dict = None) -> None:
        """Invoke command and check it failed."""
        result = self.invoke(args, settings=settings)

        cast(TestCase, self).assertNotEqual(result.exit_code, 0)
        cast(TestCase, self).assertEqual(result.stdout, stdout)
        cast(TestCase, self).assertIn(stderr, result.stderr)

==> fred-backend-messenger_0.5.0.orig.tar.gz: fred-backend-messenger-0.5.0/messenger/tests/utils.py <==
"""Utilities for tests."""
import os
import warnings
from copy import copy
from typing import Any, Dict, Sequence, Type, cast
from unittest import TestCase
from unittest.mock import ANY
from uuid import UUID

from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from click import BaseCommand
from click.testing import CliRunner, Result
from sqlalchemy import create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.event import listen, remove
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.pool import Pool
from sqlalchemy.sql import text
from sqlalchemy_utils.functions import (create_database as _create_database, database_exists as _database_exists,
                                        drop_database as _drop_database)
from testfixtures import LogCapture

from messenger.commands.migrate_archive import get_fred_engine
from messenger.constants import ArchiveMessageStatus, MessageStatus
from messenger.database import get_engine, get_session_cls
from messenger.filters import Filter
from messenger.models import (ArchiveEmail, ArchiveEmailReference, ArchiveLetter, ArchiveLetterReference, ArchiveSms,
                              ArchiveSmsReference, Email, Letter, Sms)
from messenger.settings import SETTINGS, MessengerSettings, Settings

try:
    from psycopg2.extensions import connection as psycopg2_connection
except ImportError:
    psycopg2_connection = type(None)


class TestSettings(Settings):
    """Settings manager for tests.

    Does not actually load settings from any configuration file.
    """

    def load(self, config_argv: str = None) -> None:
        """Initiate an empty settings.

        Does nothing if called multiple times.
        """
        if self._settings is None:
            self._settings = MessengerSettings()


class TestSession(Session):
    def close(self):
        if self._transaction is not None and self._transaction.is_active:
            # Flush all the prepared objects.
            self.flush()
        super().close()

    def __del__(self):
        if self._transaction is not None and self._transaction.is_active:  # pragma: no cover
            warnings.warn("The test session hasn't been closed.", ResourceWarning)


def test_session_cls():
    return sessionmaker(bind=get_engine(), class_=TestSession)


def database_exists(url):
    """Check if database exists.

    Tweaked version to use template1 database to investigate postgresql databases and use environment variables.
    """
    url = copy(make_url(url))
    if url.drivername == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if opts:
            database = url.database
        else:
            # No connection parameters, use environment instead.
            database = os.environ.get('PGDATABASE')
        url = url.set(database='template1')
        sql = text("SELECT 1 FROM pg_database WHERE datname= :database")

        engine = create_engine(url)
        try:
            return bool(engine.scalar(sql, database=database))
        finally:
            engine.dispose()
    return _database_exists(url)


def create_database(url, encoding='utf-8', template=None):
    """Create database.

    Tweaked version to use postgresql environment variables, when appropriate.
    """
    if url.drivername == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if not opts:
            # No connection parameters, use environment instead.
            url = url.set(database=os.environ.get('PGDATABASE'))
    return _create_database(url, encoding=encoding, template=template)


def drop_database(url):
    """Drop database.

    Tweaked version to use postgresql environment variables, when appropriate.
    """
    if url.drivername == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if not opts:
            # No connection parameters, use environment instead.
            url = url.set(database=os.environ.get('PGDATABASE'))
    return _drop_database(url)


def run_migrations(engine):
    """Run migrations on target database engine."""
    connection = engine.connect()
    config = Config(os.path.join(os.path.dirname(__file__), '..', 'alembic.ini'))
    scripts = ScriptDirectory.from_config(config)
    context = EnvironmentContext(config, scripts)

    def _upgrade(revision, context):
        return scripts._upgrade_revs(scripts.get_heads(), revision)

    context.configure(connection=connection, fn=_upgrade)
    context.run_migrations()
    if engine.name == 'postgresql':
        sql = "CREATE TABLE IF NOT EXISTS {table}_default PARTITION OF {table} DEFAULT"
        # Create default partitions for partitioned tables.
        engine.execute(sql.format(table=ArchiveEmail.__tablename__))
        engine.execute(sql.format(table=ArchiveEmailReference.__tablename__))
        engine.execute(sql.format(table=ArchiveSms.__tablename__))
        engine.execute(sql.format(table=ArchiveSmsReference.__tablename__))
        engine.execute(sql.format(table=ArchiveLetter.__tablename__))
        engine.execute(sql.format(table=ArchiveLetterReference.__tablename__))


def set_connection_timezone(connection, record):
    """Set time zone for postgres connection to UTC."""
    if isinstance(connection, psycopg2_connection):
        cursor = connection.cursor()
        cursor.execute('SET TIME ZONE UTC')


# XXX: Using snake_case looks much better for this context manager. It reminds override_settings from Django.
class override_settings:
    """Decorator and context manager to temporarily override settings."""

    def __init__(self, **overrides):
        self.overrides = overrides
        self.old_settings = None

    @property
    def overrides_database(self):
        """Return whether override changes database."""
        return any(n.startswith('db_') for n in self.overrides)

    def reset_database(self):
        """Reset database if required."""
        if self.overrides_database:
            get_engine.cache_clear()
            get_session_cls.cache_clear()
        if any(n.startswith('fred_db_') for n in self.overrides):
            get_fred_engine.cache_clear()

    def enable(self):
        """Start settings override."""
        self.old_settings = SETTINGS.settings
        SETTINGS.settings = TestSettings()
        # Capture logs from settings loading.
        with LogCapture('messenger.settings'):
            for key, value in dict(self.old_settings.items(), **self.overrides).items():
                SETTINGS[key] = value
        self.reset_database()
        if self.overrides_database:
            listen(Pool, 'connect', set_connection_timezone)

            # Create and migrate test database.
            if not database_exists(get_engine().url):
                create_database(get_engine().url)
            run_migrations(get_engine())

    def disable(self):
        """Stop settings override."""
        if self.overrides_database and not get_engine().name == 'sqlite':  # pragma: no cover
            # Drop the test database, if not a sqlite database.
            drop_database(get_engine().url)

            remove(Pool, 'connect', set_connection_timezone)

        SETTINGS.settings = cast(Settings, self.old_settings)
        self.reset_database()

    def __enter__(self):
        self.enable()

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()

    def __call__(self, decorated: Type[TestCase]) -> Type[TestCase]:
        """Decorate TestCase."""
        original_setup = decorated.setUp
        original_teardown = decorated.tearDown

        def setUp(inner_self):
            self.enable()
            try:
                original_setup(inner_self)
            except Exception:  # pragma: no cover
                self.disable()
                raise

        def tearDown(inner_self):
            original_teardown(inner_self)
            self.disable()

        decorated.setUp = setUp  # type: ignore
        decorated.tearDown = tearDown  # type: ignore
        return decorated


def make_email(
        session: Session, *, recipient: Sequence[str], subject_template: str, body_template: str,
        status: MessageStatus = MessageStatus.PENDING, **kwargs: Any) -> Email:
    """Make and save an email."""
    email = Email(recipient=recipient, status=status, subject_template=subject_template,
                  body_template=body_template, **kwargs)
    session.add(email)
    session.commit()
    return email


def make_archive_email(
        session: Session, *, id: int, recipient: Sequence[str], subject_template: str, body_template: str,
        status: ArchiveMessageStatus = ArchiveMessageStatus.SENT, **kwargs: Any) -> ArchiveEmail:
    """Make and save an archive email."""
    email = ArchiveEmail(id=id, recipient=recipient, status=status, subject_template=subject_template,
                         body_template=body_template, **kwargs)
    session.add(email)
    session.commit()
    return email


def make_sms(session: Session, *, recipient: str, body_template: str, status: MessageStatus = MessageStatus.PENDING,
             **kwargs: Any) -> Sms:
    """Make and save a sms."""
    sms = Sms(recipient=recipient, status=status, body_template=body_template, **kwargs)
    session.add(sms)
    session.commit()
    return sms


def make_archive_sms(session: Session, *, id: int, recipient: str, body_template: str,
                     status: ArchiveMessageStatus = ArchiveMessageStatus.SENT, **kwargs: Any) -> ArchiveSms:
    """Make and save an archive SMS."""
    sms = ArchiveSms(id=id, recipient=recipient, status=status, body_template=body_template, **kwargs)
    session.add(sms)
    session.commit()
    return sms


def make_letter(session: Session, *, file: UUID, status: MessageStatus = MessageStatus.PENDING,
                **kwargs: Any) -> Letter:
    """Make and save a letter."""
    letter = Letter(status=status, file=file, **kwargs)
    session.add(letter)
    session.commit()
    return letter


def make_archive_letter(
        session: Session, *, id: int, file: UUID, status: ArchiveMessageStatus = ArchiveMessageStatus.SENT,
        **kwargs: Any) -> ArchiveLetter:
    """Make and save an archive letter."""
    letter = ArchiveLetter(id=id, status=status, file=file, **kwargs)
    session.add(letter)
    session.commit()
    return letter


class StaticFilter(Filter):
    """Filter which always return static value.

    By default, it accept all messages, as if no filter was defined.
    """

    def __init__(self, result: bool = True):
        self.result = result

    def filter(self, model: Any) -> bool:
        # Return whatever is inserted.
        return self.result


class CommandMixin:
    """Mixin for command tests."""

    runner = CliRunner(mix_stderr=False)
    command: BaseCommand

    def invoke(self, args: Sequence[str] = None, *, settings: Dict = None) -> Result:
        """Invoke a command and return a result."""
        with override_settings(**(settings or {})):
            return self.runner.invoke(self.command, args)

    def assertCommandSuccess(self, args: Sequence[str] = None, *, stdout: str = ANY, settings: Dict = None) -> None:
        """Invoke command and check it succeeded."""
        result = self.invoke(args, settings=settings)

        err_msg = "command failed with code {}, stderr: {!r}".format(result.exit_code, result.stderr)
        cast(TestCase, self).assertEqual(result.exit_code, 0, msg=err_msg)
        cast(TestCase, self).assertEqual(result.stdout, stdout)
        cast(TestCase, self).assertEqual(result.stderr, '')

    def assertCommandFail(self, args: Sequence[str] = None, *, stdout: str = '', stderr: str = '',
                          settings: Dict = None) -> None:
        """Invoke command and check it failed."""
        result = self.invoke(args, settings=settings)

        cast(TestCase, self).assertNotEqual(result.exit_code, 0)
        cast(TestCase, self).assertEqual(result.stdout, stdout)
        cast(TestCase, self).assertIn(stderr, result.stderr)
